#################################################
#################################################
##          COPYRIGHT - DATA SOCIETY           ##
#################################################
#################################################

######################################
##    INTRODUCTION TO CLUSTERING    ##
######################################

## NOTE: To run code blocks (separated by #%%), select somewhere inside the block 
##       and press ctrl + enter for Windows or Linux, 
##       and command + enter for Macs

#==================================================================================
#%%
# Slide 27: Step 1: load data

# Load needed packages and modules
from sklearn import datasets
import numpy as np
import os
import matplotlib.pyplot as plt

# How to create dummy data for clustering
from sklearn.datasets import make_blobs

sample_data, y = make_blobs(n_samples = 150, 
                  n_features = 2, 
                  centers = 2, 
                  cluster_std = 0.5, 
                  shuffle = True, 
                  random_state = 0)

#==================================================================================
#%%
# Slide 28: Step 2: plot data as is

# sample_data is a 2D array of the coordinates of our data 
type(sample_data) 
sample_data.shape
sample_data[0:5,:] # first 5 rows of the X matrix 

# Plot of sample clusters
# Arguments in plt.scatter are the two vectors to plot.
# we style the dots to have a white fill color, and make sure to include 
# edgecolor as black, otherwise the white circles will not appear.
plt.scatter(sample_data[:, 0], sample_data[:, 1], 
            c = 'white', marker = 'o', edgecolor = 'black', s = 50)
plt.grid()
plt.title("Sample clusters")
plt.xlabel('Variable 1')
plt.ylabel('Variable 2')
plt.tight_layout()
plt.show()

plt.savefig('plots/sample_data/points_bw.png', dpi=200)
# plt.close()

#==================================================================================
#%%
# Slide 31: Step 2: run k-means

from sklearn.cluster import KMeans

# Kmeans, with parameters already set. 
km2 = KMeans(n_clusters = 2,
            init = 'random', 
            n_init = 10, 
            max_iter = 300,
            tol = 1e-04,
            random_state = 0)
# n_clusters: specify the number of clusters
# n_init = number of times we run the kmeans algorithm, choosing the final 
# model with the least SSE
# max_iter = specify the maximum number of iterations for each single run
# tol = because the kmeans algorithm will continue iterating until the 
# max_iter is reached, we set the tol parameter
#       this is the parameter that controls the tolerance with regard 
#       to the changes in the WSS to declare convergence.

# run ("fit") kmeans on data set sample_data, 
#  and categorize ("predict") a cluster for each point 
y_km2 = km2.fit_predict(sample_data)

type(y_km2)
y_km2.shape
y_km2[0:10] # look the first few values

#==================================================================================
#%%
'''
To measure/quantify the quality of the clustering, we use within cluster SSE (inertia) .
We don't have to calculate this with sklearn and the Kmeans function, which already has 
inertia_ as an object.
'''

print('Inertia: %.2f' % km2.inertia_)

#==================================================================================
#%%
# Slide 36: Step 3: get the best k value

'''
To optimize the way we choose the number of k, we can use a plot that illustrates 
the % Inertia by # of clusters.
'''

inertias = []
try_k = range(1, 11)
for k in try_k:
    km = KMeans(n_clusters = k, 
                init = 'k-means++', 
                n_init = 10, 
                max_iter = 300, 
                random_state = 0)
    km.fit(sample_data)
    inertias.append(km.inertia_)

del km

# make a line plot showing t
plt.plot(try_k, inertias, marker = 'o')
plt.xlabel('Number of clusters (k)')
plt.ylabel('Inertia')
plt.title('Elbow Plot - Optimal number of clusters', fontsize = 16)
plt.tight_layout()

plt.show()

plt.savefig('plots/sample_data/elbow.png', dpi = 200)
# plt.close()


#==================================================================================
#%%
# Slides 38 - 39: Step 4: visualize plot

# Visualize the clusters 1 & 2, as they are predicted by y_km2
cluster_0 = sample_data[y_km2 == 0, :] # <== the data points that were assigned to cluster 0
cluster_1 = sample_data[y_km2 == 1, :] # <== the data points that were assigned to cluster 1

# plot the 0 cluster
plt.scatter(cluster_0[:, 0],
            cluster_0[:, 1],
            s = 50,
            c = 'white',
            marker = 'o',
            edgecolors = 'blue',
            label = 'cluster 1')

# plot the 1 cluster
plt.scatter(cluster_1[:, 0],
            cluster_1[:, 1],
            s = 50,
            c = 'white',
            marker = 'o',
            edgecolors = 'red',
            label = 'cluster 2')

# plot the centroids as large black dots
plt.scatter(km2.cluster_centers_[:, 0],
            km2.cluster_centers_[:, 1],
            s = 250, # make the size larger for these 3
            marker = 'o',
            c = ['blue', 'red'],
            label = 'centroids')
plt.grid()
plt.title("Sample Clusters")
plt.xlabel('Variable 1')
plt.ylabel('Variable 2')

plt.savefig('plots/sample_data/kmeans_2.png', dpi = 200)
# plt.close()

#==================================================================================
#%%

############################
##    GO TO EXERCISE 1    ##
############################

#==================================================================================
#%%
'''
You may notice that within the KMeans algorithm in the for loop, 
for the init parameter we use 'kmeans++' instead of 'random'.
kmeans++ is an optimized version of kmeans which instead of randomly choosing 
centroids, it will choose centroids that are furthest
away from each other. This helps optimize the clusters and ensure the BSS is 
larger and WSS is smaller.
'''

#==================================================================================
#%%
# Slides 45: Step 5: measure clustering quality
'''
Silhouette analysis is another metric that can be used to measure the quality of the clustering. 
It can also be applied to other clustering methods besides k-means. Silhouette analysis can be used as 
a graphical tool to plot a measure of how tightly grouped the samples in the clusters are. To calculate
the silhouette coefficient, apply the following three steps:
    
    1) Calculate the cluster cohesion as the average distance between a sample and all other points in the same cluster.
    
    2) Calculate the cluster separation from the next closest cluster as the average distance between the sample and all samples
       in the nearest cluster.
       
    3) Calculate the silhouette as the difference between cluster cohesion and separation divided by the greater of the two (whichever that may be)
        
'''
       
from matplotlib import cm
from sklearn.metrics import silhouette_samples

# We are now going to create a plot of the silhouette coefficient for a k-means clustering with 2 centers, as proven 
# optimal from the elbow plot we just created in the previous slides

# We first run k-means with 2 centers
km2 = KMeans(n_clusters=2, 
            init='k-means++', 
            n_init=10, 
            max_iter=300,
            tol=1e-04,
            random_state=0)
y_km2 = km2.fit_predict(sample_data)

#==================================================================================
#%%
# Slides 46-47: Step 5: measure clustering quality

# Plot the silhouette for each cluster
# we can use 2 subplots and create a horizontal barplot on each

fig, axes = plt.subplots(nrows=2, ncols=1, sharex=True)

cluster_0 = sample_data[y_km2==0,:] # the red one
cluster_1 = sample_data[y_km2==1,:] # the blue one

silhouette_vals = silhouette_samples(sample_data, y_km2, metric='euclidean')

silhouette_0 = sorted(silhouette_vals[y_km2==0])
silhouette_1 = sorted(silhouette_vals[y_km2==1])

axes[1].barh(range(0, len(cluster_0)), silhouette_0, height=1.0, 
             edgecolor='none', color='blue')
axes[1].set_ylabel('Cluster 1')
axes[1].set_yticklabels([])

axes[0].barh(range(0, len(cluster_1)), silhouette_1, height=1.0, 
             edgecolor='none', color='red')
axes[0].set_ylabel('Cluster 2')
axes[0].set_yticklabels([])

# add a vertical line for average silhouette value
silhouette_avg = np.mean(silhouette_vals)
axes[0].axvline(silhouette_avg, color="black", linestyle="--") 
axes[1].axvline(silhouette_avg, color="black", linestyle="--") 

# eliminate the gap between the two subplots
plt.subplots_adjust(hspace=0.0)

plt.xlabel('Silhouette coefficient')
fig.suptitle("Silhouette plot for 2 Clusters")

fig.savefig('plots/sample_data/silhouette_k2.png', dpi = 200)
# fig.close()

#==================================================================================
#%%
# Slides 49-50: Step 5: measure clustering quality

'''
We will run the same code as we just did, except with three centroids.
First, let's run the clustering and plot.
'''

km3 = KMeans(n_clusters=3,  # <= Use 3 clusters
            init='k-means++',
            n_init=10,
            max_iter=300,
            tol=1e-04,
            random_state=0)

y_km3 = km3.fit_predict(sample_data)

colors = [cm.jet(float(i)/3) for i in y_km3]
plt.scatter(sample_data[:, 0], sample_data[:, 1],
            marker='o',
            c = 'white', edgecolor=colors)

centroid_colors = [cm.jet(float(i)/3) for i in range(3)]
plt.scatter(km3.cluster_centers_[:, 0], km3.cluster_centers_[:, 1],
            s=250, marker='o', c=centroid_colors, label='centroids')

plt.grid()
plt.xlabel("Variable 1")
plt.ylabel("Variable 2")
plt.title("Clustering with k = 3")
plt.tight_layout()

plt.show()

plt.savefig('plots/sample_data/kmeans_3.png', dpi = 200)
# plt.close()

# Three clusters fit the data better, but we saw when we plotted the 'inertia' values by cluster that 
# after two clusters, the elbow bends and the increase in explained variance is not large. 
# As we increase the number of clusters, the fit of the model improves,
#  but we risk **overfitting** it. Therefore it's best to follow the guidelines
#  of the elbow plot

#==================================================================================
#%%
# Slides 51-52: Step 5: measure clustering quality

# Now we get the silhouette values for each cluster
# The for loop below is looping through each of the cluster numbers, in this case i = 0, 1 and 2 (clusters 1-3).
# enumerate is a function native to Python that allows us to loop over something and have an automatic counter.
cluster_labels = np.unique(y_km3)
n_clusters = cluster_labels.shape[0]
silhouette_vals = silhouette_samples(sample_data, y_km3, metric = 'euclidean')
y_ax_lower, y_ax_upper = 0, 0
yticks = []
for i, c in enumerate(cluster_labels):
    c_silhouette_vals = silhouette_vals[y_km3 == c]
    c_silhouette_vals.sort()
    y_ax_upper += len(c_silhouette_vals)
    color = cm.jet(float(i) / n_clusters) # create a unique color based on i
    plt.barh(range(y_ax_lower, y_ax_upper), c_silhouette_vals, height=1.0, 
             edgecolor='none', color = color)

    yticks.append((y_ax_lower + y_ax_upper) / 2.)
    y_ax_lower += len(c_silhouette_vals)
    

# To summarize the goodness of our clustering, we add the average silhouette 
# coefficient line to the plot. 
silhouette_avg = np.mean(silhouette_vals)
plt.axvline(silhouette_avg, color = "red", linestyle = "--") 

plt.yticks(yticks, cluster_labels + 1)
plt.ylabel('Cluster')
plt.xlabel('Silhouette coefficient')


plt.title("Silhouette plot for 3 Clusters")
plt.tight_layout()
plt.show()

plt.savefig('plots/sample_data/silhouette_k3.png', dpi = 300)
# plt.close()

# We see a big  difference between the two clusters and the three clusters

# Compare the silhouette plot to the plot we just made of the clusters (kmeans_3.png).
#  Notice the light blue cluster is most disperse, which is reflected in the silhouette plot

#==================================================================================
#%%
# Slide 57: Step 1: read in data

# In class example: Fry's General Store

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

from sklearn.cluster import KMeans

# Step 1: read the data
candies = pd.read_csv('small_candy.csv')

# Use the Names field as the index
candies = candies.set_index('Names')
# this leaves us with only the numerical data (the various flavor scores)

# Summarize the numerical columns:
candies.describe()
# notice they are all between 0 and 1 (look at min and max)

#==================================================================================
#%%
# Slide 58: Explore the data

# Let's do some exploratory data analysis (EDA) before we do the clustering
# to see if we can find any patterns.

# Plot variables against each other:

# Plot Sweetness vs Sourness
plt.scatter(candies['Sweetness'], candies['Sourness'], 
            c = 'white', marker = 'o', 
            edgecolor = 'black', s = 50)

plt.grid()
plt.xlabel('Sweetness Rating')
plt.ylabel('Sourness Rating')
plt.title("Fry's Chocolate Bars")
plt.tight_layout()
plt.show()

plt.savefig('plots/candy/eda.png', dpi = 200)

#==================================================================================
#%%
# Slides 60-61: Step 2: select k

# Step 2: Select a good value for k (number of clusters to use)
# Notice that we run KMeans on all of the variables
#  (Sweetness, Sourness, Nuttiness, and Texture)
#  unlike in our example when we only used 2 variables.
#  KMeans can be run on any number of variables, not just 2!

# Inertia plot to assess optimal number of clusters
inertias = []
try_k = range(1, 11)
for k in try_k:
    km = KMeans(n_clusters = k, 
                init = 'k-means++', 
                n_init = 10, 
                max_iter = 300, 
                random_state = 0)
    km.fit(candies) # this will use all variables in the candies dataframe
    inertias.append(km.inertia_)
    
del km

plt.plot(try_k, inertias, marker='o')
plt.xlabel('Number of clusters')
plt.ylabel('Inertia')
plt.title('Elbow Plot - Optimal number of clusters')
plt.tight_layout()

plt.show()

plt.savefig('plots/candy/elbow.png', dpi = 200)

#==================================================================================
#%%
# Slide 63: Step 3: Run kMeans

# Kmeans, using 3 clusters, using k-means++ instead of random for centroids
km3 = KMeans(n_clusters=3, 
            init='k-means++', 
            n_init = 10, 
            max_iter = 300,
            tol=1e-04,
            random_state = 0)
# n_clusters: specify the number of clusters
# n_init = number of times we run the kmeans algorithm, choosing the final model with the least SSE
# max_iter = specify the maximum number of iterations for each single run
# tol = because the kmeans algorithm will continue iterating until the max_iter is reached, we set the tol parameter
#       this is the parameter that controls the tolerance with regard to the changes in the WSS to declare convergence.

# predict which cluster each point will fall into
y_km3 = km3.fit_predict(candies)

#==================================================================================
#%%

# Slides 64-67: Step 4: Visualize clustering results

n_clusters = 3
colors = [cm.jet(float(i) / n_clusters) for i in y_km3]

# Plot sweetness vs sourness
plt.scatter(candies['Sweetness'], candies['Sourness'],
            s = 50,
            c = 'white',
            edgecolors = colors)

plt.grid()
plt.xlabel('Sweetness')
plt.ylabel('Sourness')
plt.title("Candies Clusters")
plt.tight_layout()

plt.savefig('plots/candy/kmeans_k3_sweet_sour.png', dpi = 200)
plt.close()

#%%

# Plot sweetness vs nuttiness
plt.scatter(candies['Sweetness'], candies['Nuttiness'],
            s = 50,
            c='white',
            edgecolors=colors)

plt.grid()
plt.xlabel('Sweetness')
plt.ylabel('Nuttiness')
plt.title("Candies Clusters")
plt.tight_layout()

plt.savefig('plots/candy/kmeans_k3_sweet_nutty.png', dpi=200)
plt.close()

#%%

# Plot sweetness vs texture
plt.scatter(candies['Sweetness'], candies['Texture'],
            s = 50,
            c = 'white',
            edgecolors = colors)

plt.grid()
plt.xlabel('Sweetness')
plt.ylabel('Texture')
plt.title("Candies Clusters")
plt.tight_layout()

plt.savefig('plots/candy/kmeans_k3_sweet_texture.png', dpi=200)
plt.close()

#%%

# plot nuttiness vs texture
plt.scatter(candies['Nuttiness'], candies['Texture'],
            s = 50,
            c = 'white',
            edgecolors = colors)

plt.grid()
plt.xlabel('Nuttiness')
plt.ylabel('Texture')
plt.title("Candies Clusters")
plt.tight_layout()

plt.savefig('plots/candy/kmeans_k3_nutty_texture.png', dpi = 200)
plt.close()

#==================================================================================
#%%

############################
##    GO TO EXERCISE 2    ##
############################

#==================================================================================
#%%
# Slides 70-71: Step 5: Look at the goodness of the clusters:

# Now look at the goodness of the clusters by looking at the silhouette coefficient
from sklearn.metrics import silhouette_samples

cluster_labels = np.unique(y_km3)
n_clusters = cluster_labels.shape[0]
silhouette_vals = silhouette_samples(candies, y_km3, metric='euclidean')
y_ax_lower, y_ax_upper = 0, 0
yticks = []
for i, c in enumerate(cluster_labels):
    c_silhouette_vals = silhouette_vals[y_km3 == c]
    c_silhouette_vals.sort()
    y_ax_upper += len(c_silhouette_vals)
    color = cm.jet(float(i) / n_clusters)
    plt.barh(range(y_ax_lower, y_ax_upper), c_silhouette_vals, height=1.0, 
             edgecolor = 'none', color = color)

    yticks.append((y_ax_lower + y_ax_upper) / 2.)
    y_ax_lower += len(c_silhouette_vals)
    
silhouette_avg = np.mean(silhouette_vals)
plt.axvline(silhouette_avg, color = "red", linestyle = "--") 

plt.yticks(yticks, cluster_labels + 1)
plt.ylabel('Cluster')
plt.xlabel('Silhouette coefficient')

plt.tight_layout()

plt.show()

plt.savefig('plots/candy/silhouette_k3.png', dpi = 200)
#plt.close()

# Because this is a real world example, we are not going to get perfect clusters.
# However, we do see that all three are somewhat past the average silhouette line

#==================================================================================
#%%
# Slides 73: Application of results:

# Look at the chocolates in each cluster
candy_names = candies.index.values
candy_names[y_km3 == 0]
candy_names[y_km3 == 1]
candy_names[y_km3 == 2]

#==================================================================================
#%%
# Slides 101: AGNES: implementation in Python


# HIERARCHICAL CLUSTERING
'''
Hierarchical clustering: Agglomerative
'''
import pandas as pd
np.random.seed(123)

variables = ['X', 'Y', 'Z']
labels = ['ID_0', 'ID_1', 'ID_2', 'ID_3', 'ID_4']

# We create a random table with features X,Y and Z.
# The rows are the IDs and for each ID there is feature X,Y and Z.


X = np.random.random_sample([5, 3]) * 10
df = pd.DataFrame(X, columns = variables, index = labels)
df

#==================================================================================
#%%
# Slide 102: AGNES: implementation in Python

'''
To calculate the distance matrix as input for the hierarchical clustering algorithm,
we will use the pdist function from SciPy's spatial.distance submodule.

We are going to calculate the Euclidean distance between each pair of sample points in our 
dataset based on features X,Y and Z.
'''

from scipy.spatial.distance import pdist, squareform

row_dist = pd.DataFrame(squareform(pdist(df, metric = 'euclidean')),
                        columns = labels,
                        index = labels)

# We provide the condensed distance matrix - returned by pdist - as input.
# The squareform function creates a symmetrical matrix of pair-wise distances
row_dist

#==================================================================================
#%%
# Slide 103: AGNES: implementation in Python

'''
Now we apply the complete linkage agglomeration to our clusters using the linkage function from SciPy's 
cluster.hierarchy submodule, which returns a so-called linkage  matrix.

Let's review the function documentation, before we implement it
'''
from scipy.cluster.hierarchy import linkage
help(linkage)

''' 
Parameters
    ----------
    y : ndarray
        A condensed distance matrix. A condensed distance matrix
        is a flat array containing the upper triangular of the distance matrix.
        This is the form that ``pdist`` returns. Alternatively, a collection of
        :math:`m` observation vectors in :math:`n` dimensions may be passed as an
        :math:`m` by :math:`n` array. All elements of the condensed distance matrix
        must be finite, i.e. no NaNs or infs.
    method : str, optional
        The linkage algorithm to use. See the ``Linkage Methods`` section below
        for full descriptions.
    metric : str or function, optional
        The distance metric to use in the case that y is a collection of
        observation vectors; ignored otherwise. See the ``pdist``
        function for a list of valid distance metrics. A custom distance
        function can also be used.
    
    Returns
    -------
    Z : ndarray
        The hierarchical clustering encoded as a linkage matrix.

'''

#==================================================================================
#%%
# Slide 104: AGNES: implementation in Python

'''

We can either pass a condensed distance matrix (upper triangular) from the pdist function, or we can pass the "original" data array and define 
the metric='euclidean' argument in linkage. However, we should not pass the squareform distance matrix, which would yield different distance
values although the overall clustering could be the same.

Here are the three possible scenarios using this function:
    
*** Note: to take a closer look at the clustering results, we turn them into a pandas dataframe
'''

# Incorrect approach: Squareform distance matrix
    
row_clusters = linkage(row_dist, method = 'complete', metric = 'euclidean')
pd.DataFrame(row_clusters,
             columns = ['row label 1', 'row label 2',
                      'distance', 'no. of items in clust.'],
             index = ['cluster %d' % (i + 1)
                    for i in range(row_clusters.shape[0])])

# WARNING!
#__main__:1: ClusterWarning: scipy.cluster: The symmetric non-negative hollow 
# observation matrix looks suspiciously like an uncondensed distance matrix

#==================================================================================
#%%
# Slide 105: AGNES: implementation in Python

# Correct approach: Condensed distance matrix

row_clusters = linkage(pdist(df, metric = 'euclidean'), method = 'complete')
pd.DataFrame(row_clusters,
             columns = ['row label 1', 'row label 2',
                      'distance', 'no. of items in clust.'],
             index = ['cluster %d' % (i + 1) 
                    for i in range(row_clusters.shape[0])])

#==================================================================================
#%%
# Slide 106: AGNES: implementation in Python

# Correct approach: Input sample matrix

row_clusters = linkage(df.values, method = 'complete', metric = 'euclidean')
clusters = pd.DataFrame(row_clusters,
             columns=['row label 1', 'row label 2',
                      'distance', 'no. of items in clust.'],
             index=['cluster %d' % (i + 1)
                    for i in range(row_clusters.shape[0])])
clusters

#==================================================================================
#%%
# Slide 107: AGNES: implementation in Python

'''
Let's save the 'input sample matrix' example. We can this way refer back 
to each column:
'''

clusters[['row label 1','row label 2']]
# These two columns denote the most dissimilar members in each cluster

clusters[['distance']]
# The third column represents the distance between the dissimilar members

clusters[['no. of items in clust.']]
# The last column returns the count of the members in each cluster

#==================================================================================
#%%
# Slides 108-109: AGNES: implementation in Python

'''
Now that we have computed the correct linkage matrix, we can visualize the 
results in the form of a dendogram
'''

import matplotlib.pyplot as plt
from scipy.cluster.hierarchy import dendrogram

# make dendrogram black (part 1/2)
# from scipy.cluster.hierarchy import set_link_color_palette
# set_link_color_palette(['black'])

row_dendr = dendrogram(row_clusters, 
                       labels = labels,
                       # make dendrogram black (part 2/2)
                       # color_threshold = np.inf
                       )
plt.tight_layout()
plt.ylabel('Euclidean distance')
#plt.savefig('images/11_11.png', dpi=300, 
#            bbox_inches='tight')
plt.show()

#==================================================================================
#%%
# Slides 110: AGNES: implementation in Python

'''
We saw how to perform agglomerative hierarchical clustering using SciPy. 
However, there is also an Agglomerative Clustering implementation in 
scikit-learn which allows us to choose the number of clusters
that we want to return.

This is useful if we want to prune the hierarchical cluster tree. 
'''

from sklearn.cluster import AgglomerativeClustering

ac = AgglomerativeClustering(n_clusters = 2, 
                             affinity = 'euclidean', 
                             linkage = 'complete')
labels = ac.fit_predict(X)
print('Cluster labels: %s' % labels)

#==================================================================================
#%%

############################
##    GO TO EXERCISE 3    ##
############################

#==================================================================================




